import { type SceneState, type SceneType } from "./base"
import type { Game } from "../game"
import type { SuggestedInteraction } from "../../shared/mutations"
import type { InteractionSceneDefinition } from "../../shared/script/types"
import { getSuggestedInteractionId, type InteractionSceneEvent } from "../../shared/scene-types/interaction"
import type { SessionId } from "../session"

export const InteractionSceneType = {
    id: "interaction",
    createState(game: Game, definition: InteractionSceneDefinition): InteractionSceneState {
        return new InteractionSceneState(game, definition)
    }
} as const satisfies SceneType<InteractionSceneDefinition, InteractionSceneState>

export class InteractionSceneState implements SceneState<InteractionSceneEvent> {
    private interactionQueue: Map<string, { interaction: SuggestedInteraction; votes: Set<SessionId> }> = new Map()
    private suggestedInteractionIdBySessionId: Map<SessionId, string> = new Map()
    private ongoingInteractionExecution: SuggestedInteraction | null = null
    private objectVisibilityById = new Map<string, boolean>()
    private objectCompletionStepById = new Map<string, number>()

    constructor(private game: Game, private definition: InteractionSceneDefinition) {
        definition.objectsById.entries().forEach(([id, o]) => this.setObjectVisibility(id, o.reveal))
    }

    getConnectionEvents(): InteractionSceneEvent[] {
        const events: InteractionSceneEvent[] = []
        events.push(this.getVotesChangedEvent())

        this.objectVisibilityById.entries().forEach(([id, isVisible]) => events.push({ type: "object-visibility-changed", objectId: id, isVisible }))
        this.objectCompletionStepById.entries().forEach(([id, step]) => events.push({ type: "object-completion-step-changed", objectId: id, step }))

        if (this.ongoingInteractionExecution !== null) events.push({ type: "interaction-execution-started", interaction: this.ongoingInteractionExecution })

        return events
    }

    removePlayer(sessionId: SessionId) {
        const suggestedInteractionId = this.suggestedInteractionIdBySessionId.get(sessionId)
        if (suggestedInteractionId !== undefined) {
            const { votes } = this.interactionQueue.get(suggestedInteractionId)!
            votes.delete(sessionId)
            if (votes.size <= 0) this.interactionQueue.delete(suggestedInteractionId)

            this.emitVotesChanged()
        }

        this.suggestedInteractionIdBySessionId.delete(sessionId)
    }

    private emit(event: InteractionSceneEvent) {
        this.game.eventBus.emit("player-broadcast", { type: "scene-event", event })
    }

    private getVotesChangedEvent(): InteractionSceneEvent {
        return { type: "votes-changed", votesByInteractionId: Object.fromEntries(this.interactionQueue.entries().map(([id, { votes }]) => ([id, votes.size]))) }
    }

    private emitVotesChanged() {
        this.emit(this.getVotesChangedEvent())
    }

    setInteractionVote(interaction: SuggestedInteraction | null, sessionId: SessionId) {
        const newInteractionId = interaction === null ? null : getSuggestedInteractionId(interaction)
        const currentSuggestedInteractionId = this.suggestedInteractionIdBySessionId.get(sessionId)
        if (currentSuggestedInteractionId !== undefined) {
            if (newInteractionId === currentSuggestedInteractionId) return
            const { votes } = this.interactionQueue.get(currentSuggestedInteractionId)!
            votes.delete(sessionId)
            if (votes.size <= 0) this.interactionQueue.delete(currentSuggestedInteractionId)
            this.suggestedInteractionIdBySessionId.delete(sessionId)
        }

        if (newInteractionId !== null) {
            const existingItem = this.interactionQueue.get(newInteractionId)
            if (existingItem === undefined) {
                const item = {
                    id: newInteractionId,
                    votes: new Set([sessionId]),
                    interaction: interaction!
                }

                this.interactionQueue.set(item.id, item)
                this.emit({ type: "interaction-queued", interaction: interaction! })
            } else {
                existingItem.votes.add(sessionId)
            }

            this.suggestedInteractionIdBySessionId.set(sessionId, newInteractionId)
        }

        this.emitVotesChanged()
    }

    startInteractionExecution(interaction: SuggestedInteraction) {
        this.ongoingInteractionExecution = interaction
        this.emit({ type: "interaction-execution-started", interaction })
    }

    finishInteractionExecution(interactionId: string, onlyIfOngoing: boolean = true) {
        if (onlyIfOngoing && (this.ongoingInteractionExecution === null || interactionId !== getSuggestedInteractionId(this.ongoingInteractionExecution))) return
        this.ongoingInteractionExecution = null
        this.emit({ type: "interaction-execution-finished", interactionId })
        this.removeInteractionFromQueue(interactionId)

        const interaction = this.definition.interactionsById.get(interactionId)
        if (interaction === undefined) return

        switch (interaction.type) {
            case "use":
                if (interaction.consume) {
                    this.setObjectVisibility(interaction.objectId, false)
                }

                interaction.outputObjectIds.forEach(objectId => {
                    this.revealOrUpgradeObject(objectId, interaction.objectId === objectId && interaction.consume)
                })

                break

            case "combine":
                interaction.inputObjects.entries().forEach(([id, { consume }]) => {
                    if (consume) this.setObjectVisibility(id, false)
                })

                interaction.outputObjectIds.forEach(objectId => {
                    this.revealOrUpgradeObject(objectId, interaction.inputObjects.get(objectId)?.consume === true)
                })

                break
        }
    }

    cancelInteractionExecution(interactionId: string, onlyIfOngoing: boolean = true) {
        if (onlyIfOngoing && (this.ongoingInteractionExecution === null || interactionId !== getSuggestedInteractionId(this.ongoingInteractionExecution))) return
        this.ongoingInteractionExecution = null
        this.emit({ type: "interaction-execution-cancelled" })
    }

    removeInteractionFromQueue(id: string) {
        const queueItem = this.interactionQueue.get(id)
        if (queueItem === undefined) return
        queueItem.votes.forEach(sessionId => this.suggestedInteractionIdBySessionId.delete(sessionId))
        this.interactionQueue.delete(id)
        this.emitVotesChanged()
    }

    setObjectVisibility(objectId: string, isVisible: boolean) {
        const current = this.objectVisibilityById.get(objectId)
        if (current === isVisible) return

        this.objectVisibilityById.set(objectId, isVisible)
        this.emit({ type: "object-visibility-changed", objectId: objectId, isVisible })
    }

    revealOrUpgradeObject(objectId: string, resetCompletionSteps: boolean = false) {
        const object = this.definition.objectsById.get(objectId)!
        this.setObjectVisibility(objectId, true)

        if (object.completion !== undefined) {
            let step = (this.objectCompletionStepById.get(objectId) ?? -1)
            if (resetCompletionSteps) {
                step = 0
            } else {
                step = Math.min(step + 1, object.completion.steps)
            }

            this.objectCompletionStepById.set(objectId, step)
            this.emit({ type: "object-completion-step-changed", objectId, step })

            if (step === object.completion.steps) {
                this.objectVisibilityById.set(objectId, false)
                this.objectVisibilityById.set(object.completion.replaceWith, true)
            }
        }
    }
}