import type { SceneController, SceneType } from "../base"
import type { InteractionSceneDefinition } from "../../../shared/script/types"
import PlayerView from "./PlayerView.vue"
import type { InteractionSceneEvent } from "../../../shared/scene-types/interaction"
import { getSuggestedInteractionId } from "../../../shared/scene-types/interaction"
import CrewView from "./CrewView.vue"
import type { SuggestedInteraction } from "../../../shared/mutations"
import { computed, reactive, type Reactive, ref, type Ref } from "vue"
import { trpcClient } from "../../trpc"

export const InteractionSceneType: SceneType<InteractionSceneDefinition, InteractionSceneController> = {
    id: "interaction",
    playerView: PlayerView,
    crewView: CrewView,
    createController(definition: InteractionSceneDefinition): InteractionSceneController {
        const visibleObjectIds = reactive(new Set<string>())
        const objectCompletionById = reactive(new Map<string, number>())
        const replacedObjectIdMap = reactive(new Map<string, string>())
        const interactionQueue = reactive(new Map<string, InteractionQueueItem>())

        const suggestedInteraction = ref<SuggestedInteraction | null>(null)
        const suggestedInteractionId = computed(() => suggestedInteraction.value === null ? null : getSuggestedInteractionId(suggestedInteraction.value))

        const ongoingInteractionExecution = ref<SuggestedInteraction | null>(null)
        const ongoingInteractionExecutionId = computed(() => ongoingInteractionExecution.value === null ? null : getSuggestedInteractionId(ongoingInteractionExecution.value))

        return {
            visibleObjectIds,
            objectCompletionById,
            interactionQueue,
            sortedInteractionQueue: computed(() => [...interactionQueue.values()].sort((a, b) => a.votes - b.votes)),
            sortedInteractionQueueWithDefined: computed(() => {
                const result: InteractionQueueItem[] = [...interactionQueue.values()].sort((a, b) => a.votes - b.votes)

                for (const [interactionId, interaction] of definition.interactionsById.entries()) {
                    let suggestion: SuggestedInteraction
                    switch (interaction.type) {
                        case "use":
                            suggestion = {
                                type: "use",
                                objectId: interaction.objectId
                            }

                            break

                        case "combine":
                            suggestion = {
                                type: "combine",
                                objectIds: new Set(interaction.inputObjects.keys())
                            }
                    }

                    if (!interactionQueue.has(interactionId)) {
                        result.push({
                            id: interactionId,
                            interaction: suggestion,
                            votes: 0
                        })
                    }
                }

                return result
            }),
            suggestedInteraction,
            suggestedInteractionId,
            ongoingInteractionExecution,
            ongoingInteractionExecutionId,
            handleEvent(event: InteractionSceneEvent) {
                switch (event.type) {
                    case "votes-changed":
                        const removedInteractionIds = new Set(interactionQueue.keys())
                        Object.entries(event.votesByInteractionId).forEach(([interactionId, votes]) => {
                            if (votes > 0) {
                                interactionQueue.get(interactionId)!.votes = votes
                                removedInteractionIds.delete(interactionId)
                            }
                        })

                        removedInteractionIds.forEach(interactionId => {
                            if (interactionId === suggestedInteractionId.value) suggestedInteraction.value = null
                            return interactionQueue.delete(interactionId)
                        })

                        break

                    case "interaction-queued":
                        const id = getSuggestedInteractionId(event.interaction)
                        interactionQueue.set(id, { id, interaction: event.interaction, votes: 1 })
                        break

                    case "object-visibility-changed":
                        if (event.isVisible) visibleObjectIds.add(event.objectId)
                        else visibleObjectIds.delete(event.objectId)
                        break

                    case "object-completion-step-changed":
                        objectCompletionById.set(event.objectId, event.step)
                        const object = definition.objectsById.get(event.objectId)!
                        if (object.completion!.steps === event.step) {
                            replacedObjectIdMap.set(event.objectId, object.completion!.replaceWith)
                            visibleObjectIds.delete(event.objectId)
                            visibleObjectIds.add(object.completion!.replaceWith)
                        }

                        break

                    case "interaction-execution-started":
                        ongoingInteractionExecution.value = event.interaction
                        break

                    case "interaction-execution-cancelled":
                        ongoingInteractionExecution.value = null
                        break

                    case "interaction-execution-finished":
                        ongoingInteractionExecution.value = null
                        interactionQueue.delete(event.interactionId)
                        if (suggestedInteractionId.value === event.interactionId) suggestedInteraction.value = null
                        break
                }
            },
            async setObjectVisibility(objectId: string, isVisible: boolean) {
                await trpcClient.crew.interactionScene.setObjectVisibility.mutate({ objectId, isVisible })
            },
            async setInteractionVote(interaction: SuggestedInteraction | null) {
                await trpcClient.player.interactionScene.setInteractionVote.mutate({ interaction })
                suggestedInteraction.value = interaction
            },
            async startInteractionExecution(interaction: SuggestedInteraction) {
                await trpcClient.crew.interactionScene.startInteractionExecution.mutate({ interaction })
            },
            async cancelInteractionExecution(interactionId: string, onlyIfOngoing: boolean) {
                await trpcClient.crew.interactionScene.cancelInteractionExecution.mutate({ interactionId, onlyIfOngoing })
            },
            async finishInteractionExecution(interactionId: string, onlyIfOngoing: boolean) {
                await trpcClient.crew.interactionScene.finishInteractionExecution.mutate({ interactionId, onlyIfOngoing })
            },
        }
    }
}

export interface InteractionQueueItem {
    id: string
    interaction: SuggestedInteraction
    votes: number
}

export interface InteractionSceneController extends SceneController {
    visibleObjectIds: Reactive<Set<string>>
    interactionQueue: Reactive<Map<string, InteractionQueueItem>>
    objectCompletionById: Reactive<Map<string, number>>
    sortedInteractionQueue: Readonly<Ref<Array<InteractionQueueItem>>>
    sortedInteractionQueueWithDefined: Readonly<Ref<Array<InteractionQueueItem>>>
    suggestedInteraction: Ref<SuggestedInteraction | null>
    suggestedInteractionId: Readonly<Ref<string | null>>
    ongoingInteractionExecution: Ref<SuggestedInteraction | null>
    ongoingInteractionExecutionId: Readonly<Ref<string | null>>

    setObjectVisibility(objectId: string, isVisible: boolean): Promise<void>

    setInteractionVote(interaction: SuggestedInteraction | null): Promise<void>

    startInteractionExecution(interaction: SuggestedInteraction): Promise<void>

    cancelInteractionExecution(interactionId: string, onlyIfOngoing: boolean): Promise<void>

    finishInteractionExecution(interactionId: string, onlyIfOngoing: boolean): Promise<void>
}