import { Temporal } from "temporal-polyfill"
import { getSuggestedInteractionFromDefinition, getSuggestedInteractionId } from "../scene-types/interaction"
import { cMap, cSet } from "../util"

export interface Script {
    scenesById: Map<string, SceneDefinition>
}

export interface SceneDefinitionBase {
    id: string
    type: string
    label: string
    plannedDuration: Temporal.Duration | null
}

export const defineInteractionScene = <ObjectsT extends Record<string, SceneObjectDefinition<keyof ObjectsT & string>>>(d: SceneDefinitionBase & {
    type: "interaction"
    objects: ObjectsT
    interactions: Array<RawUseInteractionDefinition<keyof ObjectsT & string> | RawCombineInteractionDefinition<keyof ObjectsT & string>>
}): InteractionSceneDefinition<keyof ObjectsT & string> => ({
    ...d,
    objectsById: cMap(d.objects),
    interactionsById: new Map(d.interactions.values().map(raw => {
        let i: UseInteractionDefinition<keyof ObjectsT & string> | CombineInteractionDefinition<keyof ObjectsT & string>
        switch (raw.type) {
            case "use":
                i = {
                    type: "use",
                    objectId: raw.objectId,
                    consume: raw.consume,
                    outputObjectIds: cSet(...raw.outputObjectIds),
                    note: raw.note
                }
                break

            case "combine":
                i = {
                    type: "combine",
                    inputObjects: cMap(raw.inputObjects),
                    outputObjectIds: cSet(...raw.outputObjectIds),
                    note: raw.note
                }
        }
        return [getSuggestedInteractionId(getSuggestedInteractionFromDefinition(i)), i]
    }))
})

export interface InteractionSceneDefinition<ObjectIds extends string = string> extends SceneDefinitionBase {
    type: "interaction"
    objectsById: Map<ObjectIds, SceneObjectDefinition<ObjectIds>>
    interactionsById: Map<string, UseInteractionDefinition<ObjectIds> | CombineInteractionDefinition<ObjectIds>>
}

export interface SceneObjectDefinition<AvailableObjectIds extends string = string> {
    label: string
    reveal: boolean
    completion?: SceneObjectDefinitionCompletionOptions<AvailableObjectIds>
}

export interface SceneObjectDefinitionCompletionOptions<AvailableObjectIds extends string = string> {
    steps: number
    replaceWith: AvailableObjectIds
}

interface RawUseInteractionDefinition<AvailableObjectIds extends string> {
    type: "use"
    objectId: AvailableObjectIds
    consume: boolean
    outputObjectIds: Array<AvailableObjectIds>
    note?: string
}

export interface UseInteractionDefinition<AvailableObjectIds extends string> {
    type: "use"
    objectId: AvailableObjectIds
    consume: boolean
    outputObjectIds: Set<AvailableObjectIds>
    note?: string
}

interface RawCombineInteractionDefinition<AvailableObjectIds extends string> {
    type: "combine"
    inputObjects: Partial<Record<AvailableObjectIds, CombinationInteractionDefinitionInputOptions>>
    outputObjectIds: Array<AvailableObjectIds>
    note?: string
}

export interface CombineInteractionDefinition<AvailableObjectIds extends string> {
    type: "combine"
    inputObjects: Map<AvailableObjectIds, CombinationInteractionDefinitionInputOptions>
    outputObjectIds: Set<AvailableObjectIds>
    note?: string
}

export interface CombinationInteractionDefinitionInputOptions {
    consume: boolean
}

export interface ChoiceSceneDefinition extends SceneDefinitionBase {
    type: "choice"
    options: ChoiceSceneDefinitionOption[]
}

export interface ChoiceSceneDefinitionOption {
    id: string
    label: string
}

export interface TextSceneDefinition extends SceneDefinitionBase {
    type: "text"
    text: string
}

export type SceneDefinition = InteractionSceneDefinition | ChoiceSceneDefinition | TextSceneDefinition
